# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

from pyrasta.base import RasterBase
from pyrasta.tools.dem import _slope, _aspect


class Raster(RasterBase):
    pass


class DigitalElevationModel(Raster):

    def aspect(self, scale=1):
        """ Compute DEM aspect

        Parameters
        ----------
        scale: float or int
            Ratio of vertical units to horizontal

        Returns
        -------

        """
        return _aspect(self, scale)

    def slope(self, slope_format="percent", scale=1):
        """ Compute DEM slope

        Parameters
        ----------
        slope_format: str
            Slope format {'percent', 'degree'}
        scale: int or float
            Ratio of vertical units to horizontal

        Returns
        -------

        """
        return _slope(self, slope_format, scale)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

from functools import wraps

import numpy as np
from pyrasta.io_.files import RasterTempFile
from pyrasta.tools import _gdal_temp_dataset
from sklearn.cluster import KMeans

import gdal


def return_classification(classification):
    @wraps(classification)
    def _return_classification(raster, nb_classes, *args, **kwargs):
        with RasterTempFile(raster._gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:
            out_ds = _gdal_temp_dataset(out_file.path, raster._gdal_driver,
                                        raster._gdal_dataset.GetProjection(),
                                        raster.x_size, raster.y_size, 1,
                                        raster.geo_transform, gdal.GetDataTypeByName('Int16'),
                                        no_data=-999)
            labels = classification(raster, nb_classes, out_ds, *args, **kwargs)

            # Close dataset
            out_ds = None

            new_raster = raster.__class__(out_file.path)
            new_raster._temp_file = out_file

        return new_raster
    return _return_classification


@return_classification
def _k_means_classification(raster, nb_clusters, out_ds, n_init, max_iter):
    """ Apply k-means classification

    """
    k_means_classifier = KMeans(nb_clusters, n_init=n_init, max_iter=max_iter)
    values = np.reshape(raster._gdal_dataset.ReadAsArray(),
                        (raster.nb_band, raster.x_size * raster.y_size)).transpose()
    labels = np.reshape(k_means_classifier.fit(values).labels_, (raster.y_size, raster.x_size))

    out_ds.GetRasterBand(1).WriteArray(labels)


def k_means_classification(raster, nb_clusters, n_init=100, max_iter=1000):
    _k_means_classification(raster, nb_clusters, n_init, max_iter)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from collections import Collection
from itertools import chain, islice

from numba import njit
from tqdm import tqdm


MP_CHUNK_SIZE = 1000


class TqdmUpTo(tqdm):
    """ Progress bar for url retrieving

    Description
    -----------
    Thanks to https://github.com/tqdm/tqdm/blob/master/examples/tqdm_wget.py

    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.
    Inspired by [twine#242](https://github.com/pypa/twine/pull/242),
    [here](https://github.com/pypa/twine/commit/42e55e06).
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)  # also sets self.n = b * bsize


def check_string(string, list_of_strings):
    """ Check validity of and return string against list of valid strings

    :param string: searched string
    :param list_of_strings: list/tuple/set of valid strings string is to be checked against
    :return: validate string from list of strings if match
    """
    check_type(string, str, list_of_strings, Collection)
    [check_type(x, str) for x in list_of_strings]

    output_string = []

    for item in list_of_strings:
        if item.lower().startswith(string.lower()):
            output_string.append(item)

    if len(output_string) == 1:
        return output_string[0]
    elif len(output_string) == 0:
        raise ValueError("input must match one of those: {}".format(list_of_strings))
    elif len(output_string) > 1:
        raise ValueError("input match more than one valid value among {}".format(list_of_strings))


def check_type(*args):
    """Check type of arguments

    :param args: tuple list of argument/type
    :return:
    """
    if len(args) % 2 == 0:
        for item in range(0, len(args), 2):
            if not isinstance(args[item], args[item + 1]):
                raise TypeError("Type of argument {} is '{}' but must be '{}'".format(
                    item//2 + 1, type(args[item]).__name__, args[item + 1].__name__))


def digitize(value, list_of_values, ascend=True, right=False):
    """

    Description
    -----------

    Parameters
    ----------

    """
    if ascend:
        loc = [value > v if right else value >= v for v in list_of_values]
    else:
        loc = [value <= v if right else value < v for v in list_of_values]

    return loc.index(False) if False in loc else len(list_of_values)


@njit()
def grid(origin, resolution, size):
    """ Return regular grid vector

    Parameters
    ----------
    origin
    resolution
    size

    Returns
    -------

    """
    origin -= resolution

    for _ in range(size):
        origin += resolution

        yield origin


def lazyproperty(func):
    name = '_lazy_' + func.__name__

    @property
    def lazy(self):
        if hasattr(self, name):
            return getattr(self, name)
        else:
            value = func(self)
            setattr(self, name, value)
            return value
    return lazy


def split_into_chunks(iterable, size):
    """ Split iterable into chunks of iterables

    """
    iterator = iter(iterable)
    for first in iterator:
        yield chain([first], islice(iterator, size - 1))



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
import numpy as np

from pyrasta.io_.files import RasterTempFile
from pyrasta.tools import _gdal_temp_dataset, _return_raster
from pyrasta.tools.windows import get_block_windows, get_xy_block_windows
from tqdm import tqdm

import gdal


@_return_raster
def _op(raster1, out_file, raster2, op_type):
    """ Basic arithmetic operations

    """
    out_ds = _gdal_temp_dataset(out_file, raster1._gdal_driver,
                                raster1._gdal_dataset.GetProjection(), raster1.x_size,
                                raster1.y_size, raster1.nb_band, raster1.geo_transform,
                                gdal.GetDataTypeByName('Float32'), raster1.no_data)

    for band in range(1, raster1.nb_band + 1):

        for window in get_block_windows(1000, raster1.x_size, raster1.y_size):
            array1 = raster1._gdal_dataset.GetRasterBand(
                band).ReadAsArray(*window).astype("float32")
            try:
                array2 = raster2._gdal_dataset.GetRasterBand(
                    band).ReadAsArray(*window).astype("float32")
            except AttributeError:
                array2 = raster2  # If second input is not a raster but a scalar

            if op_type == "add":
                result = array1 + array2
            elif op_type == "sub":
                result = array1 - array2
            elif op_type == "mul":
                result = array1 * array2
            elif op_type == "truediv":
                result = array1 / array2
            else:
                result = None

            out_ds.GetRasterBand(band).WriteArray(result, window[0], window[1])

    # Close dataset
    out_ds = None


def _raster_calculation(raster_class, sources, fhandle, window_size,
                        gdal_driver, data_type, no_data, nb_processes,
                        chunksize):
    """ Calculate raster expression

    """
    if not hasattr(window_size, "__getitem__"):
        window_size = (window_size, window_size)

    master_raster = sources[0]
    with RasterTempFile(gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:

        is_first_run = True
        y = 0
        pgbar = tqdm(total=master_raster.y_size // window_size[1] + int(master_raster.y_size %
                                                                        window_size[1] != 0),
                     desc="Calculate raster expression")

        while y < master_raster.y_size:

            y_size = min(window_size[1], master_raster.y_size - y)

            arrays = []
            for src in sources:
                array = src._gdal_dataset.ReadAsArray(0, y, None, y_size).astype("float32")
                array[array == src.no_data] = np.nan
                arrays.append(array)

            window_gen = ([array[w[1]:w[1] + w[3], w[0]:w[0] + w[2]] for array in arrays] for w
                          in get_xy_block_windows(window_size, master_raster.x_size, y_size))

            with mp.Pool(processes=nb_processes) as pool:
                result = np.concatenate(list(pool.imap(fhandle,
                                                       window_gen,
                                                       chunksize=chunksize)),
                                        axis=1)

            result[np.isnan(result)] = no_data

            pgbar.update(1)

            if is_first_run:
                if result.ndim == 2:
                    nb_band = 1
                else:
                    nb_band = result.shape[0]

                out_ds = _gdal_temp_dataset(out_file.path,
                                            gdal_driver,
                                            master_raster._gdal_dataset.GetProjection(),
                                            master_raster.x_size,
                                            master_raster.y_size, nb_band,
                                            master_raster.geo_transform,
                                            data_type,
                                            no_data)

                is_first_run = False

            if nb_band == 1:
                out_ds.GetRasterBand(1).WriteArray(result, 0, y)
            else:
                for band in range(nb_band):
                    out_ds.GetRasterBand(band + 1).WriteArray(result[band, :, :],
                                                              0, y)

            y += window_size[1]

    # Close dataset
    out_ds = None

    return raster_class(out_file.path)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from pyrasta.crs import srs_from
from pyrasta.io_.files import RasterTempFile, VrtTempFile
from pyrasta.tools import _gdal_temp_dataset, _return_raster

import affine
import gdal


@_return_raster
def _align_raster(in_raster, out_file, on_raster):
    """ Align raster on other raster

    """
    out_ds = _gdal_temp_dataset(out_file, in_raster._gdal_driver,
                                on_raster._gdal_dataset.GetProjection(),
                                on_raster.x_size, on_raster.y_size, in_raster.nb_band,
                                on_raster.geo_transform, in_raster.data_type, in_raster.no_data)

    gdal.Warp(out_ds, in_raster._gdal_dataset)

    # Close dataset
    out_ds = None


@_return_raster
def _extract_bands(raster, out_file, bands):

    out_ds = gdal.Translate(out_file, raster._gdal_dataset, bandList=bands)

    # Close dataset
    out_ds = None


def _xy_to_2d_index(raster, x, y):
    """ Convert x/y map coordinates to 2d index

    """
    forward_transform = affine.Affine.from_gdal(*raster.geo_transform)
    reverse_transform = ~forward_transform
    px, py = reverse_transform * (x, y)

    return int(px), int(py)


def _merge_bands(raster_class, sources, resolution, gdal_driver, data_type, no_data):
    """ Merge multiple bands into one raster

    """
    with RasterTempFile(gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:

        vrt_ds = gdal.BuildVRT(VrtTempFile().path, [src._gdal_dataset for src in sources],
                               resolution=resolution, separate=True, VRTNodata=no_data)
        out_ds = gdal.Translate(out_file.path, vrt_ds, outputType=data_type)

    # Close dataset
    out_ds = None

    return raster_class(out_file.path)


@_return_raster
def _padding(raster, out_file, pad_x, pad_y, pad_value):
    """ Add pad values around raster

    Description
    -----------

    Parameters
    ----------
    raster: RasterBase
        raster to pad
    out_file: str
        output file to which to write new raster
    pad_x: int
        x padding size (new width will therefore be RasterXSize + 2 * pad_x)
    pad_y: int
        y padding size (new height will therefore be RasterYSize + 2 * pad_y)
    pad_value: int or float
        value to set to pad area around raster

    Returns
    -------
    """
    geo_transform = (raster.x_origin - pad_x * raster.resolution[0], raster.resolution[0], 0,
                     raster.y_origin + pad_y * raster.resolution[1], 0, -raster.resolution[1])
    out_ds = _gdal_temp_dataset(out_file,
                                raster._gdal_driver,
                                raster._gdal_dataset.GetProjection(),
                                raster.x_size + 2 * pad_x,
                                raster.y_size + 2 * pad_y,
                                raster.nb_band,
                                geo_transform,
                                raster.data_type,
                                raster.no_data)

    for band in range(1, raster.nb_band + 1):
        out_ds.GetRasterBand(band).Fill(pad_value)
        gdal.Warp(out_ds, raster._gdal_dataset)

    # Close dataset
    out_ds = None


@_return_raster
def _project_raster(raster, out_file, new_crs):
    """ Project raster onto new CRS

    """
    gdal.Warp(out_file, raster._gdal_dataset, dstSRS=srs_from(new_crs))


def _read_array(raster, band, bounds):
    """ Read array from raster

    """
    if bounds is None:
        return raster._gdal_dataset.ReadAsArray()
    else:
        x_min, y_min, x_max, y_max = bounds
        forward_transform = affine.Affine.from_gdal(*raster.geo_transform)
        reverse_transform = ~forward_transform
        px_min, py_max = reverse_transform * (x_min, y_min)
        px_max, py_min = reverse_transform * (x_max, y_max)
        x_size = int(px_max - px_min) + 1
        y_size = int(py_max - py_min) + 1

        if band is not None:
            return raster._gdal_dataset.GetRasterBand(band).ReadAsArray(int(px_min),
                                                                        int(py_min),
                                                                        x_size,
                                                                        y_size)
        else:
            return raster._gdal_dataset.ReadAsArray(int(px_min),
                                                    int(py_min),
                                                    x_size,
                                                    y_size)


def _read_value_at(raster, x, y):
    """ Read value at lat/lon map coordinates

    """
    forward_transform = affine.Affine.from_gdal(*raster.geo_transform)
    reverse_transform = ~forward_transform
    xoff, yoff = reverse_transform * (x, y)
    value = raster._gdal_dataset.ReadAsArray(xoff, yoff, 1, 1)
    if value.size > 1:
        return value
    else:
        return value[0, 0]


@_return_raster
def _resample_raster(raster, out_file, factor):
    """ Resample raster

    Parameters
    ----------
    raster: RasterBase
        raster to resample
    out_file: str
        output file to which to write new raster
    factor: int or float
        Resampling factor
    """
    geo_transform = (raster.x_origin, raster.resolution[0] / factor, 0,
                     raster.y_origin, 0, -raster.resolution[1] / factor)
    out_ds = _gdal_temp_dataset(out_file,
                                raster._gdal_driver,
                                raster._gdal_dataset.GetProjection(),
                                raster.x_size * factor,
                                raster.y_size * factor,
                                raster.nb_band,
                                geo_transform,
                                raster.data_type,
                                raster.no_data)

    for band in range(1, raster.nb_band+1):
        gdal.RegenerateOverview(raster._gdal_dataset.GetRasterBand(band),
                                out_ds.GetRasterBand(band), 'mode')

    # Close dataset
    out_ds = None


@_return_raster
def _rescale_raster(raster, out_file, ds_min, ds_max):

    out_ds = gdal.Translate(out_file, raster._gdal_dataset,
                            scaleParams=[[src_min, src_max, ds_min, ds_max]
                                         for src_min, src_max in zip(raster.min, raster.max)])

    # Close dataset
    out_ds = None



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

__version__ = '1.2.4'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2020, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from functools import wraps, partial

from numba import jit
from tqdm import tqdm

import multiprocessing as mp
import numpy as np

from pyrasta.tools import _gdal_temp_dataset, _return_raster
from pyrasta.exceptions import WindowGeneratorError
from pyrasta.utils import split_into_chunks, check_string, check_type, MP_CHUNK_SIZE


def _set_nan(array, function, no_data):
    """ Replace no data values by NaNs

    """
    array[array == no_data] = np.nan
    return function(array)


@_return_raster
def _windowing(raster, out_file, function, band, window_size,
               method, data_type, no_data, chunk_size, nb_processes):
    """ Apply function in each moving or block window in raster

    Description
    -----------

    Parameters
    ----------

    """
    window_generator = WindowGenerator(raster, band, window_size, method)
    out_ds = _gdal_temp_dataset(out_file, raster._gdal_driver, raster._gdal_dataset.GetProjection(),
                                window_generator.x_size, window_generator.y_size, raster.nb_band,
                                window_generator.geo_transform, data_type, no_data)

    y = 0
    # chunk size cannot be 0 and cannot
    # be higher than height of window
    # generator (y_size). And it must be
    # a multiple of window generator width
    # (x_size)
    chunk_size = max(min(chunk_size // window_generator.x_size, window_generator.y_size)
                     * window_generator.x_size, window_generator.x_size)
    for win_gen in tqdm(split_into_chunks(window_generator, chunk_size),
                        total=len(window_generator)//chunk_size +
                        int(len(window_generator) % chunk_size != 0),
                        desc="Sliding window computation"):
        with mp.Pool(processes=nb_processes) as pool:
            output = np.asarray(list(pool.map(partial(_set_nan,
                                                      function=function,
                                                      no_data=raster.no_data),
                                              win_gen,
                                              chunksize=MP_CHUNK_SIZE)))

        output[np.isnan(output)] = no_data

        # Set number of rows to write to file
        n_rows = len(output) // window_generator.x_size

        # Write row to raster
        out_ds.GetRasterBand(band).WriteArray(np.reshape(output, (n_rows,
                                                                  window_generator.x_size)), 0, y)

        # Update row index
        y += n_rows

    # Close dataset
    out_ds = None


def integer(setter):

    @wraps(setter)
    def _integer(self, value):
        try:
            check_type(value, int)
        except TypeError:
            raise WindowGeneratorError("'%s' must be an integer value but is: '%s'" %
                                       (setter.__name__, type(value).__name__))
        output = setter(self, value)

    return _integer


def odd(setter):

    @wraps(setter)
    def _odd(self, value):
        if value % 2 == 0:
            raise WindowGeneratorError("'%s' must be an odd value (=%d)" % (setter.__name__, value))
        output = setter(self, value)

    return _odd


def positive(setter):

    @wraps(setter)
    def _positive(self, value):
        if value <= 0:
            raise WindowGeneratorError("'%s' must be positive (=%d)" % (setter.__name__, value))
        output = setter(self, value)

    return _positive


class WindowGenerator:
    """ Generator of windows over raster

    """

    def __init__(self, raster, band, window_size, method):
        """ WindowGenerator constructor

        Description
        -----------

        Parameters
        ----------
        raster: RasterBase
            raster for which we must compute windows
        band: int
            raster band number
        window_size: int
            size of window in pixels
        method: str
            sliding window method ("block" or "moving")

        Return
        ------

        """
        self.band = band
        self.raster = raster
        self.window_size = window_size
        self.method = method

        # self.image = self.raster._gdal_dataset.GetRasterBand(self.band).ReadAsArray()

    @property
    def geo_transform(self):
        if self.method == "block":
            topleftx, pxsizex, rotx, toplefty, roty, pxsizey = \
                self.raster._gdal_dataset.GetGeoTransform()
            return topleftx, pxsizex * self.window_size, rotx, \
                toplefty, roty, pxsizey * self.window_size
        else:
            return self.raster._gdal_dataset.GetGeoTransform()

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        try:
            self._method = check_string(value, {'block', 'moving'})
        except (TypeError, ValueError) as e:
            raise WindowGeneratorError("Invalid sliding window method: '%s'" % value)

    @property
    def band(self):
        return self._band

    @band.setter
    @integer
    @positive
    def band(self, value):
        self._band = value

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    @integer
    @positive
    def window_size(self, value):
        self._window_size = value

    @property
    def x_size(self):
        if self.method == "block":
            return int(self.raster.x_size / self.window_size) + \
                   min(1, self.raster.x_size % self.window_size)
        else:
            return self.raster.x_size

    @property
    def y_size(self):
        if self.method == "block":
            return int(self.raster.y_size / self.window_size) + \
                   min(1, self.raster.y_size % self.window_size)
        else:
            return self.raster.y_size

    def __len__(self):
        return self.y_size * self.x_size

    def __iter__(self):
        def windows():
            if self.method == "block":
                return get_block_windows(self.window_size, self.raster.x_size, self.raster.y_size)
            elif self.method == "moving":
                return get_moving_windows(self.window_size, self.raster.x_size, self.raster.y_size)

        return (self.raster._gdal_dataset.GetRasterBand(self.band).ReadAsArray(*window).astype(
            "float32") for window in windows())
        # return (self.image[w[1]:w[1] + w[3], w[0]:w[0] + w[2]] for w in windows())


@jit(nopython=True, nogil=True)
def get_xy_block_windows(window_size, raster_x_size, raster_y_size):
    """ Get xy block window coordinates

    Description
    -----------
    Get xy block window coordinates depending
    on raster size and window size

    Parameters
    ----------
    window_size: (int, int)
        size of window to read within raster as (width, height)
    raster_x_size: int
        raster's width
    raster_y_size: int
        raster's height

    Yields
    -------
    Window coordinates: tuple
        4-element tuple returning the coordinates of the window within the raster
    """
    for y in range(0, raster_y_size, window_size[1]):
        ysize = min(window_size[1], raster_y_size - y)
        for x in range(0, raster_x_size, window_size[0]):
            xsize = min(window_size[0], raster_x_size - x)

            yield x, y, xsize, ysize


@jit(nopython=True, nogil=True)
def get_block_windows(window_size, raster_x_size, raster_y_size):
    """ Get block window coordinates

    Description
    -----------
    Get block window coordinates depending
    on raster size and window size

    Parameters
    ----------
    window_size: int
        size of window to read within raster
    raster_x_size: int
        raster's width
    raster_y_size: int
        raster's height

    Yields
    -------
    Window coordinates: tuple
        4-element tuple returning the coordinates of the window within the raster
    """
    for y in range(0, raster_y_size, window_size):
        ysize = min(window_size, raster_y_size - y)
        for x in range(0, raster_x_size, window_size):
            xsize = min(window_size, raster_x_size - x)

            yield x, y, xsize, ysize


@jit(nopython=True, nogil=True)
def get_moving_windows(window_size, raster_x_size, raster_y_size, step=1):
    """ Get moving window coordinates

    Description
    -----------
    Get moving window coordinates depending
    on raster size, window size and step

    Parameters
    ----------
    window_size: int
        size of window (square)
    raster_x_size: int
        raster's width
    raster_y_size: int
        raster's height
    step: int
        gap between the window's centers when moving the window over the raster

    Yields
    -------
    Window coordinates: tuple
        tuple of coordinates
    """
    offset = int((window_size - 1) / 2)  # window_size must be an odd number
    # for each pixel, compute indices of the window (all included)
    for y in range(0, raster_y_size, step):
        y1 = max(0, y - offset)
        y2 = min(raster_y_size - 1, y + offset)
        ysize = (y2 - y1) + 1
        for x in range(0, raster_x_size, step):
            x1 = max(0, x - offset)
            x2 = min(raster_x_size - 1, x + offset)
            xsize = (x2 - x1) + 1

            yield x1, y1, xsize, ysize



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from zipfile import ZipFile
from urllib.error import URLError
from urllib.request import urlretrieve
import tempfile
import os

from tqdm import tqdm

from pyrasta.raster import DigitalElevationModel
from pyrasta.utils import digitize, TqdmUpTo

import gdal


CGIAR_ARCHIVE_FORMAT = "zip"
CGIAR_URL = "http://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF"
CGIAR_NO_DATA = -32768
CGIAR_DATA_TYPE = gdal.GetDataTypeByName('Int16')


def _download_srtm_tile(tile_name):
    """ Download and extract SRTM tile archive

    Description
    -----------

    Parameters
    ----------
    tile_name: str
        SRTM tile name
    """
    zip_name, tif_name = tile_name + "." + CGIAR_ARCHIVE_FORMAT, tile_name + '.tif'
    url = os.path.join(CGIAR_URL, zip_name)
    temp_srtm_zip = os.path.join(tempfile.gettempdir(), zip_name)
    temp_srtm_dir = os.path.join(tempfile.gettempdir(), tile_name)

    # Download tile
    try:
        with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1,
                      desc=zip_name) as t:
            urlretrieve(url, temp_srtm_zip, reporthook=t.update_to)
            t.total = t.n
    except URLError as e:
        raise RuntimeError("Unable to fetch data at '%s': %s" % (url, e))

    # Extract GeoTiff
    # In order to avoid conflict with the archive extraction,
    # the "gdal import" is located at the end of all the imports
    # See https://github.com/conda-forge/gdal-feedstock/issues/365
    with ZipFile(temp_srtm_zip, 'r') as archive:
        archive.extractall(temp_srtm_dir)

    return os.path.join(temp_srtm_dir, tif_name)


def _retrieve_cgiar_srtm_tiles(bounds):
    """ Import DEM tile from CGIAR-CSI SRTM3 database (V4.1)

    Description
    -----------

    Parameters
    ----------
    bounds: tuple or list
        output DEM bounds as (x_min, y_min, x_max, y_max)

    Returns
    -------
    list:
        list of SRTM tile file names

    """
    srtm_lon = range(-180, 185, 5)
    srtm_lat = range(60, -65, -5)
    x_min = digitize(bounds[0], srtm_lon, right=False)
    x_max = digitize(bounds[2], srtm_lon, right=True)
    y_min = digitize(bounds[3], srtm_lat, right=True, ascend=False)
    y_max = digitize(bounds[1], srtm_lat, right=False, ascend=False)

    list_of_tiles = []

    coords = [(x, y) for x in range(int(x_min), int(x_max) + 1)
              for y in range(int(y_min), int(y_max) + 1)]

    for (x, y) in tqdm(coords, desc="Downloading SRTM tile(s)"):
        tile = _download_srtm_tile("srtm_%02d_%02d" % (x, y))
        list_of_tiles.append(tile)

    return list_of_tiles


def from_cgiar_online_database(bounds):
    """ Build DEM tile from CGIAR-CSI SRTM3 database (V4.1)

    Description
    -----------

    Parameters
    ----------
    bounds: tuple or list
        output DEM bounds

    Returns
    -------
    DigitalElevationModel:
        new instance

    """
    tiles = [DigitalElevationModel(tile) for tile in _retrieve_cgiar_srtm_tiles(bounds)]

    return DigitalElevationModel.merge(tiles,
                                       bounds,
                                       data_type=CGIAR_DATA_TYPE,
                                       no_data=CGIAR_NO_DATA)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
import numpy as np
import osmnx as ox
from gistools.layer import PolygonLayer
from matplotlib import pyplot

from pyrasta.raster import DigitalElevationModel, Raster
from pyrasta.tools.clip import _clip_raster_by_mask
from pyrasta.tools.rasterize import _rasterize
from pyrasta.tools.srtm import from_cgiar_online_database

# bounds = (5, 40, 10, 45)
# test = from_cgiar_online_database(bounds)
# test.to_file("/home/benjamin/srtm_corse.tif")

# test = DigitalElevationModel("/home/benjamin/dem_test.tif").clip((13, 42, 15, 45)).to_file(
#     "/home/benjamin/dem_clip_test.tif")
# pop = Raster("/home/benjamin/Documents/PRO/PRODUITS/POPULATION_DENSITY/001_DONNEES/COTE_D_IVOIRE"
#              "/population_civ_2019-07-01_geotiff/population_civ_2019-07-01.tif")
from pyrasta.tools.stats import _zonal_stats

dem = DigitalElevationModel("/home/benjamin/Documents/PRO/PRODUITS/TESTS/dem_ci.tif")
test = dem.to_crs(32630).slope("degree")
test.to_file("/home/benjamin/dem_slope_ci.tif")

country = PolygonLayer.from_gpd(ox.geocode_to_gdf(
    dict(country="Cote d'Ivoire",
         admin_level=2,
         type="boundary"))).to_crs(32630).clean_geometry()

# dem = from_cgiar_online_database(country.total_bounds)
# dem.to_file("/home/benjamin/dem_ci.tif")

honeycomb = country.split(country.area[0] / 100, method="hexana", show_progressbar=True)

honeycomb = honeycomb.to_crs(dem.crs)
honeycomb.to_file("/home/benjamin/Documents/PRO/PRODUITS/TESTS/honeycomb.shp")
honeycomb["ID"] = honeycomb.index

# test = dem.clip(mask=honeycomb[[15]], all_touched=True)
# test.to_file("/home/benjamin/pop.tif")

# test = dem.zonal_stats(honeycomb, stats=["mean", "median"])

# test = dem.rasterize(honeycomb, dem.projection, dem.x_size, dem.y_size, dem.geo_transform,
#                      attribute="ID", nb_band=1)
# test = _rasterize(DigitalElevationModel, honeycomb, None, "ID", dem._gdal_driver,
#                   dem._gdal_dataset.GetProjection(),
#                   dem.x_size, dem.y_size, 1, dem.geo_transform, dem.data_type, -999, True)
# test = DigitalElevationModel(test.path)
# test.to_file("/home/benjamin/test_ci.tif")

stats = _zonal_stats(dem, honeycomb, 1, ['mean', 'median', 'max', 'min'], dict(mymax=np.max), -999,
                     True, True, 6)

print(stats.keys())
print("mymax: %s" % stats["mymax"])
print(len(honeycomb))



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from pyrasta.io_.files import RasterTempFile

import gdal


def _merge(raster_class, sources, bounds, output_format, data_type, no_data):
    """ Merge multiple raster sources

    Description
    -----------

    Parameters
    ----------
    sources: list
        list of RasterBase instances
    """

    # Extent of all inputs
    if bounds is not None:
        dst_w, dst_s, dst_e, dst_n = bounds
    else:
        # scan input files
        xs = [item for src in sources for item in src.bounds[0:2]]
        ys = [item for src in sources for item in src.bounds[2::]]
        dst_w, dst_s, dst_e, dst_n = min(xs), min(ys), max(xs), max(ys)

    with RasterTempFile(gdal.GetDriverByName(output_format).GetMetadata()['DMD_EXTENSION']) \
            as out_file:
        gdal.Warp(out_file.path, [src._gdal_dataset for src in sources],
                  outputBounds=(dst_w, dst_s, dst_e, dst_n),
                  format=output_format, srcNodata=[src.no_data for src in sources],
                  dstNodata=no_data, outputType=data_type)

    # Be careful with the temp file, make a pointer to be sure
    # the Python garbage collector does not destroy it !
    raster = raster_class(out_file.path)
    raster._temp_file = out_file

    return raster


# def _rasterio_merge_modified(sources, out_file, bounds=None, driver="GTiff", precision=7):
#     """ Modified rasterio merge
#
#     Description
#     -----------
#     Merge set of rasters using modified
#     rasterio merging tool. Only import as
#     numpy arrays the source datasets, and
#     write each source to destination raster.
#
#     Parameters
#     ----------
#     sources: list
#         list of rasterio datasets
#     out_file: str
#         valid path to the raster file to be written
#     bounds: tuple
#         valid boundary tuple (optional)
#     driver: str
#         valid gdal driver (optional)
#     precision: int
#         float precision (optional)
#
#     Note
#     ----
#     Adapted from
#     https://gis.stackexchange.com/questions/348925/merging-rasters-with-rasterio-in-blocks-to-avoid-memoryerror
#     """
#
    # adapted from https://github.com/mapbox/rasterio/blob/master/rasterio/merge.py
    # first = sources[0]
    # first_res = first.res
    # dtype = first.dtypes[0]
    # Determine output band count
    # output_count = first.count

    # Extent of all inputs
    # if bounds:
    #     dst_w, dst_s, dst_e, dst_n = bounds
    # else:
        # scan input files
        # xs = []
        # ys = []
        # for src in sources:
        #     left, bottom, right, top = src.bounds
        #     xs.extend([left, right])
        #     ys.extend([bottom, top])
        # dst_w, dst_s, dst_e, dst_n = min(xs), min(ys), max(xs), max(ys)
    #
    # out_transform = Affine.translation(dst_w, dst_n)
    #
    # Resolution/pixel size
    # res = first_res
    # out_transform *= Affine.scale(res[0], -res[1])
    #
    # Compute output array shape. We guarantee it will cover the output
    # bounds completely
    # output_width = int(np.ceil((dst_e - dst_w) / res[0]))
    # output_height = int(np.ceil((dst_n - dst_s) / res[1]))

    # Adjust bounds to fit
    # dst_e, dst_s = out_transform * (output_width, output_height)

    # create destination array
    # destination array shape
    # shape = (output_height, output_width)
    #
    # dest_profile = {
    #     "driver": driver,
    #     "height": shape[0],
    #     "width": shape[1],
    #     "count": output_count,
    #     "dtype": dtype,
    #     "crs": sources[0].crs.to_proj4(),
    #     "nodata": sources[0].nodata,
    #     "transform": out_transform
    # }

    # open output file in write/read mode and fill with destination mosaick array
    # with rasterio.open(out_file, 'w+', **dest_profile) as mosaic_raster:
    #     for src in sources:

            # 1. Compute spatial intersection of destination and source
            # src_w, src_s, src_e, src_n = src.bounds
            # int_w = src_w if src_w > dst_w else dst_w
            # int_s = src_s if src_s > dst_s else dst_s
            # int_e = src_e if src_e < dst_e else dst_e
            # int_n = src_n if src_n < dst_n else dst_n

            # 2. Compute the source window
            # src_window = windows.from_bounds(
            #     int_w, int_s, int_e, int_n, src.transform, precision=precision)
            #
            # src_window = src_window.round_shape()

            # 3. Compute the destination window
            # dst_window = windows.from_bounds(
            #     int_w, int_s, int_e, int_n, out_transform, precision=precision)
            # dst_window = windows.Window(int(round(dst_window.col_off)),
            #                             int(round(dst_window.row_off)),
            #                             int(round(dst_window.width)),
            #                             int(round(dst_window.height)))
            #
            # out_shape = (dst_window.height, dst_window.width)

            # for band in range(1, output_count+1):
            #     src_array = src.read(band, out_shape=out_shape, window=src_window)
                # before writing the window, replace source nodata with dest nodata as it can
                # already have been written (e.g. another adjacent country)
                # dst_array = mosaic_raster.read(band, out_shape=out_shape, window=dst_window)
                # mask = src_array == src.nodata
                # src_array[mask] = dst_array[mask]
                # mosaic_raster.write(src_array, band, window=dst_window)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

from functools import wraps

from pyrasta.io_.files import RasterTempFile

import gdal


def _return_raster(function):
    @wraps(function)
    def return_raster(raster, *args, **kwargs):
        with RasterTempFile(raster._gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:
            function(raster, out_file.path, *args, **kwargs)
            new_raster = raster.__class__(out_file.path)
            new_raster._temp_file = out_file

        return new_raster
    return return_raster


def _gdal_temp_dataset(out_file, gdal_driver, projection, x_size, y_size,
                       nb_band, geo_transform, data_type, no_data):
    """ Create gdal temporary dataset

    """
    try:
        out_ds = gdal_driver.Create(out_file, x_size, y_size, nb_band, data_type)
    except RuntimeError:
        out_ds = gdal.GetDriverByName('Gtiff').Create(out_file, x_size,
                                                      y_size, nb_band, data_type)

    out_ds.SetGeoTransform(geo_transform)
    out_ds.SetProjection(projection)
    _set_no_data(out_ds, no_data)

    return out_ds


def _set_no_data(gdal_ds, no_data):
    """ Set no data value into gdal dataset

    Description
    -----------

    Parameters
    ----------
    gdal_ds: gdal.Dataset
        gdal dataset
    no_data: list or tuple
        list of no data values corresponding to each raster band

    """
    for band in range(gdal_ds.RasterCount):
        try:
            gdal_ds.GetRasterBand(band + 1).SetNoDataValue(no_data)
        except TypeError:
            pass



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
import warnings

from pyrasta.crs import proj4_from
from pyrasta.io_.files import _copy_to_file
from pyrasta.tools.calculator import _op, _raster_calculation
from pyrasta.tools.clip import _clip_raster_by_extent, _clip_raster_by_mask
from pyrasta.tools.conversion import _resample_raster, _padding, _rescale_raster, \
    _align_raster, _extract_bands, _merge_bands, _read_array, _xy_to_2d_index, _read_value_at, \
    _project_raster
from pyrasta.exceptions import RasterBaseError
from pyrasta.tools.merge import _merge
from pyrasta.tools.rasterize import _rasterize
from pyrasta.tools.stats import _histogram, _zonal_stats
from pyrasta.tools.windows import _windowing
from pyrasta.utils import lazyproperty, grid

import gdal

gdal.UseExceptions()


class RasterBase:

    def __init__(self, src_file, no_data=None):
        """ Raster class constructor

        Description
        -----------

        Parameters
        ----------
        src_file: str
            valid path to raster file
        no_data: int or float
            Set no data value only if it is not already defined in raster file
        """
        try:
            self._gdal_dataset = gdal.Open(src_file)
        except RuntimeError as e:
            raise RasterBaseError('\nGDAL returns: \"%s\"' % e)

        # If NoData not defined, define here
        for band in range(self.nb_band):
            if no_data is not None:
                if self._gdal_dataset.GetRasterBand(band + 1).GetNoDataValue() is None:
                    self._gdal_dataset.GetRasterBand(band + 1).SetNoDataValue(no_data)
                else:
                    warnings.warn("No data value is already set, cannot overwrite.")

        self._gdal_driver = self._gdal_dataset.GetDriver()
        self._file = src_file

    def __add__(self, other):
        """ Add two raster

        """
        return _op(self, other, "add")

    def __sub__(self, other):
        """ Subtract two raster

        """
        return _op(self, other, "sub")

    def __mul__(self, other):
        """ Multiply two raster

        """
        return _op(self, other, "mul")

    def __truediv__(self, other):
        """ Divide two raster

        """
        return _op(self, other, "truediv")

    def __del__(self):
        self._gdal_dataset = None

    def align_raster(self, other):
        """ Align raster on other

        Description
        -----------

        Parameters
        ----------
        other: RasterBase
            other RasterBase instance

        """

        return _align_raster(self, other)

    def clip(self, bounds=None, mask=None, no_data=-999, all_touched=True):
        """ Clip raster

        Parameters
        ----------
        bounds: tuple
            tuple (x_min, y_min, x_max, y_max) in map units
        mask: geopandas.GeoDataFrame
            Valid mask layer
        no_data: int or float
            No data value
        all_touched: bool
            if True, all touched pixels within layer boundaries are burnt,
            when clipping raster by mask

        Returns
        -------
        RasterBase:
            New temporary instance

        """
        if bounds is not None:
            return _clip_raster_by_extent(self, bounds, no_data)
        elif mask is not None:
            return _clip_raster_by_mask(self, mask, no_data, all_touched)
        else:
            raise ValueError("Either bounds or mask must be set")

    def extract_bands(self, bands):
        """ Extract bands as multiple rasters

        Description
        -----------

        Parameters
        ----------
        bands: list
            list of band numbers
        """
        return _extract_bands(self, bands)

    def histogram(self, nb_bins=10, normalized=True):
        """ Compute raster histogram

        Description
        -----------

        Parameters
        ----------
        nb_bins: int
            number of bins for histogram
        normalized: bool
            if True, normalize histogram frequency values

        Returns
        -------

        """
        return _histogram(self, nb_bins, normalized)

    @classmethod
    def merge(cls, rasters, bounds=None, output_format="Gtiff",
              data_type=gdal.GetDataTypeByName('Float32'), no_data=-999):
        """ Merge multiple rasters

        Description
        -----------

        Parameters
        ----------
        rasters: Collection
            Collection of RasterBase instances
        bounds: tuple
            bounds of the new merged raster
        output_format:str
            raster file output format (Gtiff, etc.)
        data_type: int
            GDAL data type
        no_data: int or float
            output no data value in merged raster

        Returns
        -------

        """
        return _merge(cls, rasters, bounds, output_format, data_type, no_data)

    @classmethod
    def merge_bands(cls, rasters, resolution="highest",
                    gdal_driver=gdal.GetDriverByName("Gtiff"),
                    data_type=gdal.GetDataTypeByName('Float32'),
                    no_data=-999):
        """ Create one single raster from multiple bands

        Description
        -----------
        Create one raster from multiple bands using gdal

        Parameters
        ----------
        rasters: Collection
            Collection of RasterBase instances
        resolution: str
            GDAL resolution option ("highest", "lowest", "average")
        gdal_driver: osgeo.gdal.Driver
        data_type: int
            GDAL data type
        no_data: int or float
            no data value in output raster
        """
        return _merge_bands(cls, rasters, resolution, gdal_driver, data_type, no_data)

    def pad_extent(self, pad_x, pad_y, value):
        """ Pad raster extent with given values

        Description
        -----------
        Pad raster extent, i.e. add pad value around raster bounds

        Parameters
        ----------
        pad_x: int
            x padding size (new width will therefore be RasterXSize + 2 * pad_x)
        pad_y: int
            y padding size (new height will therefore be RasterYSize + 2 * pad_y)
        value: int or float
            value to set to pad area around raster

        Returns
        -------
        RasterBase
            A padded RasterBase
        """
        return _padding(self, pad_x, pad_y, value)

    @classmethod
    def rasterize(cls, layer, projection, x_size, y_size, geo_transform,
                  burn_values=None, attribute=None,
                  gdal_driver=gdal.GetDriverByName("Gtiff"), nb_band=1,
                  data_type=gdal.GetDataTypeByName("Float32"), no_data=-999,
                  all_touched=True):
        """ Rasterize geographic layer

        Parameters
        ----------
        layer: geopandas.GeoDataFrame or gistools.layer.GeoLayer
            Geographic layer to be rasterized
        projection: str
            Projection as a WKT string
        x_size: int
            Raster width
        y_size: int
            Raster height
        geo_transform: tuple
        burn_values: list[float] or list[int], default None
            List of values to be burnt in each band, excusive with attribute
        attribute: str, default None
            Layer's attribute to be used for values to be burnt in raster,
            excusive with burn_values
        gdal_driver: osgeo.gdal.Driver, default GeoTiff
            GDAL driver
        nb_band: int, default 1
            Number of bands
        data_type: int, default "Float32"
            GDAL data type
        no_data: int or float, default -999
            No data value
        all_touched: bool

        Returns
        -------

        """
        return _rasterize(cls, layer, burn_values, attribute, gdal_driver, projection,
                          x_size, y_size, nb_band, geo_transform, data_type, no_data,
                          all_touched)

    @classmethod
    def raster_calculation(cls, rasters, fhandle, window_size=100,
                           gdal_driver=gdal.GetDriverByName("Gtiff"),
                           data_type=gdal.GetDataTypeByName('Float32'),
                           no_data=-999, nb_processes=mp.cpu_count(),
                           chunksize=1):
        """ Raster expression calculation

        Description
        -----------
        Calculate raster expression stated in "fhandle"
        such as: fhandle(raster1, raster2, etc.)
        Calculation is made for each band.

        Parameters
        ----------
        rasters: list or tuple
            collection of RasterBase instances
        fhandle: function
            expression to calculate (must accept a collection of arrays)
        window_size: int or (int, int)
            size of window/chunk to set in memory during calculation
        gdal_driver: osgeo.gdal.Driver
            GDAL driver (output format)
        data_type: int
            GDAL data type for output raster
        no_data: int or float
            no data value in resulting raster
        nb_processes: int
            number of processes for multiprocessing pool
        chunksize: int
            chunk size used in map/imap multiprocessing function

        Returns
        -------
        RasterBase
            New temporary instance
        """
        return _raster_calculation(cls, rasters, fhandle, window_size,
                                   gdal_driver, data_type, no_data,
                                   nb_processes, chunksize)

    def read_array(self, band=None, bounds=None):
        """ Write raster to numpy array

        Parameters
        ----------
        band: int
            Band number. If None, read all bands into multidimensional array.
        bounds: tuple
            tuple as (x_min, y_min, x_max, y_max) in map units. If None, read
            the whole raster into array

        Returns
        -------
        numpy.ndarray

        """
        return _read_array(self, band, bounds)

    def read_value_at(self, x, y):
        """ Read value in raster at x/y map coordinates

        Parameters
        ----------
        x: float
            lat coordinates in map units
        y: float
            lon coordinates in map units

        Returns
        -------

        """
        return _read_value_at(self, x, y)

    def resample(self, factor):
        """ Resample raster

        Description
        -----------
        Resample raster with respect to resampling factor.
        The higher the factor, the higher the resampling.

        Parameters
        ----------
        factor: int or float
            Resampling factor

        Returns
        -------
        RasterBase
            New temporary resampled instance
        """
        return _resample_raster(self, factor)

    def rescale(self, r_min, r_max):
        """ Rescale values from raster

        Description
        -----------

        Parameters
        ----------
        r_min: int or float
            minimum value of new range
        r_max: int or float
            maximum value of new range

        Returns
        -------
        """
        return _rescale_raster(self, r_min, r_max)

    def to_crs(self, crs):
        """ Re-project raster onto new CRS

        Parameters
        ----------
        crs: int or str
            valid CRS (Valid EPSG code, valid proj string, etc.)

        Returns
        -------

        """
        return _project_raster(self, crs)

    def to_file(self, filename):
        """ Write raster copy to file

        Description
        -----------
        Write raster to given file

        Parameters
        ----------
        filename: str
            File path to write to

        Return
        ------
        """
        return _copy_to_file(self, filename)

    def windowing(self, f_handle, window_size, method, band=None,
                  data_type=gdal.GetDataTypeByName('Float32'),
                  no_data=None, chunk_size=100000, nb_processes=mp.cpu_count()):
        """ Apply function within sliding/block window

        Description
        -----------

        Parameters
        ----------
        f_handle: function
        window_size: int
            size of window
        method: str
            sliding window method ('block' or 'moving')
        band: int
            raster band
        data_type: int
            gdal data type
        no_data: list or tuple
            raster no data
        chunk_size: int
            data chunk size for multiprocessing
        nb_processes: int
            number of processes for multiprocessing

        Return
        ------
        RasterBase
            New instance

        """
        if band is None:
            band = 1

        if no_data is None:
            no_data = self.no_data

        return _windowing(self, f_handle, band, window_size, method,
                          data_type, no_data, chunk_size, nb_processes)

    def xy_to_2d_index(self, x, y):
        """ Convert x/y map coordinates into 2d index

        Parameters
        ----------
        x: float
            x coordinates in map units
        y: float
            y coordinates in map units

        Returns
        -------
        tuple
            (px, py) index

        """
        return _xy_to_2d_index(self, x, y)

    def zonal_stats(self, layer, band=1, stats=None, customized_stats=None,
                    no_data=-999, all_touched=True, show_progressbar=True,
                    nb_processes=mp.cpu_count()):
        """ Compute zonal statistics

        Compute statistic among raster values
        within each feature of given geographic layer

        Parameters
        ----------
        layer: geopandas.GeoDataFrame or gistools.layer.GeoLayer
            Geographic layer
        band: int
            Band number
        stats: list[str]
            list of valid statistic names
            "mean", "median", "min", "max", "sum", "std"
        customized_stats: dict
            User's own customized statistic functions
            as {'your_function_name': function}
        no_data: int or float
            No data value
        all_touched: bool
            Whether to include every raster cell touched by a geometry, or only
            those having a center point within the polygon.
        show_progressbar: bool
            If True, show progress bar status
        nb_processes: int
            number of processes for multiprocessing

        Returns
        -------
        dict[list]
            Dictionary with each statistic as a list corresponding
            to the values for each feature in layer

        """
        if stats is not None:
            return _zonal_stats(self, layer, band, stats, customized_stats,
                                no_data, all_touched, show_progressbar, nb_processes)

    @property
    def crs(self):
        """ Return Coordinate Reference System

        """
        return proj4_from(self._gdal_dataset.GetProjection())

    @lazyproperty
    def bounds(self):
        """ Return raster bounds

        """
        return self.x_origin, self.y_origin - self.resolution[1] * self.y_size, \
            self.x_origin + self.resolution[0] * self.x_size, self.y_origin

    @lazyproperty
    def geo_transform(self):
        return self._gdal_dataset.GetGeoTransform()

    @lazyproperty
    def grid_y(self):
        return [lat for lat in grid(self.y_origin + self.geo_transform[5]/2,
                                    self.geo_transform[5], self.y_size)]

    @lazyproperty
    def grid_x(self):
        return [lon for lon in grid(self.x_origin + self.geo_transform[1]/2,
                                    self.geo_transform[1], self.x_size)]

    @lazyproperty
    def max(self):
        """ Return raster maximum value for each band

        """
        return [self._gdal_dataset.GetRasterBand(band + 1).ComputeRasterMinMax()[1]
                for band in range(self.nb_band)]

    @lazyproperty
    def mean(self):
        """ Compute raster mean for each band

        """
        return [self._gdal_dataset.GetRasterBand(band + 1).ComputeStatistics(False)[2]
                for band in range(self.nb_band)]

    @lazyproperty
    def min(self):
        """ Return raster minimum value for each band

        """
        return [self._gdal_dataset.GetRasterBand(band + 1).ComputeRasterMinMax()[0]
                for band in range(self.nb_band)]

    @lazyproperty
    def nb_band(self):
        """ Return raster number of bands

        """
        return self._gdal_dataset.RasterCount

    @lazyproperty
    def no_data(self):
        # return [self._gdal_dataset.GetRasterBand(band + 1).GetNoDataValue()
        # for band in range(self.nb_band)]
        return self._gdal_dataset.GetRasterBand(1).GetNoDataValue()

    @lazyproperty
    def data_type(self):
        return self._gdal_dataset.GetRasterBand(1).DataType

    @lazyproperty
    def resolution(self):
        """ Return raster X and Y resolution

        """
        return self.geo_transform[1], abs(self.geo_transform[5])

    @lazyproperty
    def std(self):
        """ Compute raster standard deviation for each band

        """
        return [self._gdal_dataset.GetRasterBand(band + 1).ComputeStatistics(False)[3]
                for band in range(self.nb_band)]

    @lazyproperty
    def projection(self):
        """ Get projection as a WKT string

        """
        return self._gdal_dataset.GetProjection()

    @lazyproperty
    def x_origin(self):
        return self.geo_transform[0]

    @lazyproperty
    def x_size(self):
        return self._gdal_dataset.RasterXSize

    @lazyproperty
    def y_origin(self):
        return self.geo_transform[3]

    @lazyproperty
    def y_size(self):
        return self._gdal_dataset.RasterYSize



# -*- coding: utf-8 -*-

""" Digital Elevation Model functions

More detailed description.
"""
from pyrasta.tools import _return_raster

import gdal


@_return_raster
def _slope(dem, out_file, slope_format, scale):
    """ Compute DEM slope

    Parameters
    ----------
    dem: pyrasta.raster.DigitalElevationModel
    out_file: str
        output file path to which new dem must be written
    slope_format: str
        Slope format {'percent', 'degree'}

    Returns
    -------

    """
    options = gdal.DEMProcessingOptions(format=dem._gdal_driver.ShortName,
                                        slopeFormat=slope_format,
                                        scale=scale)
    gdal.DEMProcessing(out_file, dem._gdal_dataset, 'slope', options=options)


@_return_raster
def _aspect(dem, out_file, scale):
    """ Compute aspect

    Parameters
    ----------
    dem
    out_file
    scale

    Returns
    -------

    """
    options = gdal.DEMProcessingOptions(format=dem._gdal_driver.ShortName,
                                        scale=scale)
    gdal.DEMProcessing(out_file, dem._gdal_dataset, "aspect", options=options)
